@@ -1,36 +1,39 @@
-from __future__ import division
+from __future__ import annotations
 
 from collections import defaultdict
 from itertools import count
+from typing import Optional, TextIO, ValuesView
+
+from dark.reads import Read
 
 from midtools.distances import DistanceCache
 from midtools.offsets import OffsetBases
+from midtools.read import AlignedRead
 from midtools.utils import nucleotidesToStr, s
 
 
-class ReadCluster(object):
+class ReadCluster:
     """
     Maintain a cluster of reads.
     """
 
     MIN_COMMONEST_MULTIPLE = 10.0
 
-    def __init__(self):
-        self.nucleotides = defaultdict(OffsetBases)
-        self.reads = set()
+    def __init__(self) -> None:
+        self.nucleotides: dict[int, OffsetBases] = defaultdict(OffsetBases)
+        self.reads: set[AlignedRead] = set()
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.reads)
 
-    def __str__(self):
-        result = ['Cluster with %d read%s:' %
-                  (len(self.reads), s(len(self.reads)))]
+    def __str__(self) -> str:
+        result = ["Cluster with %d read%s:" % (len(self.reads), s(len(self.reads)))]
         for read in self.reads:
-            result.append('  %s' % read)
-        result.append(nucleotidesToStr(self.nucleotides, prefix='  '))
-        return '\n'.join(result)
+            result.append("  %s" % read)
+        result.append(nucleotidesToStr(self.nucleotides, prefix="  "))
+        return "\n".join(result)
 
-    def add(self, read):
+    def add(self, read: AlignedRead) -> None:
         """
         Add a read to this cluster.
 
@@ -43,7 +46,16 @@ class ReadCluster(object):
         for offset, base in read.significantOffsets.items():
             nucleotides[offset].incorporateBase(base)
 
-    def merge(self, other):
+    def update(self, reads: list[AlignedRead]) -> None:
+        """
+        Add reads to this cluster.
+
+        @param reads: An iterable of C{AlignedRead} instances.
+        """
+        for read in reads:
+            self.add(read)
+
+    def merge(self, other: ReadCluster) -> None:
         """
         Merge another cluster into this one.
 
@@ -56,8 +68,19 @@ class ReadCluster(object):
         for offset, offsetBases in other.nucleotides.items():
             self.nucleotides[offset].merge(offsetBases)
 
+    @property
+    def offsets(self) -> set[int]:
+        """
+        Get the set of significant offsets covered by the reads in this component.
+
+        @return: A C{set} of C{int} site offsets.
+        """
+        # This is a property to match the one in the Component class in
+        # clusterAnalysis.py
+        return set(self.nucleotides)
+
     @staticmethod
-    def commonOffsetsMaxFraction(a, b):
+    def commonOffsetsMaxFraction(a: ReadCluster, b: ReadCluster) -> float:
         """
         Compute the fraction of a cluster's sites that are in common with
         another cluster. Do this for two clusters and return the higher
@@ -79,49 +102,50 @@ class ReadCluster(object):
         return len(common) / min(len(a.nucleotides), len(b.nucleotides))
 
     @staticmethod
-    def commonNucleotidesMultiplicativeDistance(a, b):
+    def commonNucleotidesMultiplicativeDistance(
+        a: ReadCluster, b: ReadCluster
+    ) -> float:
         """
-        Measure the distance from one cluster to another, as the sum of the
-        multiplied probabilities of nucleotides.
+        Measure the distance from one cluster to another, as 1.0 minus
+        the average similarity across all common offsets.
 
         @param a: A C{ReadCluster} instance.
         @param b: A C{ReadCluster} instance.
-        @raise ZeroDivisionError: if C{a} or C{b} has no offsets (neither of
-            which should be possible in normal operation).
+        @raise ZeroDivisionError: if either C{a} or C{b} have no offsets
+            (neither of which should be possible in normal operation).
         @return: The C{float} [0.0, 1.0] distance between C{a} and C{b}.
         """
-        aNucleotides = a.nucleotides
-        bNucleotides = b.nucleotides
-        commonOffsets = set(aNucleotides) & set(bNucleotides)
+        aNt = a.nucleotides
+        bNt = b.nucleotides
+        commonOffsets = set(aNt) & set(bNt)
 
         if commonOffsets:
-            similarity = sum(
-                # 1.0 - OffsetBases.multiplicativeDistance(
-                #        aNucleotides[offset], bNucleotides[offset])
-                1.0 - min(
-                    OffsetBases.multiplicativeDistance(
-                        aNucleotides[offset], bNucleotides[offset]),
-                    OffsetBases.homogeneousDistance(
-                        aNucleotides[offset], bNucleotides[offset]))
-                for offset in commonOffsets)
-            return 1.0 - (similarity / len(commonOffsets))
+            totalSimilarity = sum(
+                1.0
+                - min(
+                    OffsetBases.multiplicativeDistance(aNt[offset], bNt[offset]),
+                    OffsetBases.homogeneousDistance(aNt[offset], bNt[offset]),
+                )
+                for offset in commonOffsets
+            )
+            return 1.0 - (totalSimilarity / len(commonOffsets))
         else:
             return 1.0
 
     @staticmethod
-    def commonNucleotidesAgreementDistance(a, b):
+    def commonNucleotidesAgreementDistance(a: ReadCluster, b: ReadCluster) -> float:
         """
         Measure the distance from one cluster to another.
 
         The distance is 1.0 minus the fraction of common sites that either
-        (a) agree on what the most common nucleotide is a the site or else
+        (a) agree on what the most common nucleotide is at the site or else
         (b) where one cluster has an overwhelming opinion about the most
-        likely nucleotide.
+            likely nucleotide.
 
             In (a), each site in both clusters is examined for its most
             frequent nucleotide set (there may be more than one equally
             frequent nucleotide). If the intersection of the two sets for a
-            site is non-empty, that site counts as matching. E.g. if the
+            site is non-empty, that site counts as matching. E.g., if the
             cluster for one site has 6 x A and 2 x C and the same site in
             the other cluster has 3 x A and 3 x G, the sites agree because
             A is in the most common nucleotides set for each cluster.
@@ -142,22 +166,22 @@ class ReadCluster(object):
         @return: The C{float} [0.0, 1.0] distance between C{a} and C{b}.
 
         """
-        aNucleotides = a.nucleotides
-        bNucleotides = b.nucleotides
-        commonOffsets = set(aNucleotides) & set(bNucleotides)
+        aNt = a.nucleotides
+        bNt = b.nucleotides
+        commonOffsets = set(aNt) & set(bNt)
 
         if commonOffsets:
             matching = 0
             for offset in commonOffsets:
-                aNucleotidesAtOffset = aNucleotides[offset]
-                bNucleotidesAtOffset = bNucleotides[offset]
-                if (aNucleotidesAtOffset.commonest &
-                        bNucleotidesAtOffset.commonest):
+                aNtAtOffset = aNt[offset]
+                bNtAtOffset = bNt[offset]
+                if aNtAtOffset.commonest & bNtAtOffset.commonest:
                     # This is case (a) above.
                     matching += 1
                 else:
                     multiple = OffsetBases.highestFrequenciesMultiple(
-                        aNucleotidesAtOffset, bNucleotidesAtOffset)
+                        aNtAtOffset, bNtAtOffset
+                    )
                     # Sanity: the multiple cannot be None because that
                     # would mean only one nucleotide is present, and that
                     # case is dealt with by the first part of this if/then.
@@ -171,21 +195,20 @@ class ReadCluster(object):
             return 1.0
 
 
-class ReadClusters(object):
+class ReadClusters:
     """
-    Maintain clusters of reads.
+    Maintain multiple clusters of reads.
     """
 
     COMMON_OFFSETS_MAX_FRACTION_MIN = 0.9
 
-    def __init__(self):
+    def __init__(self) -> None:
         self._count = count()
         self.readClusters = defaultdict(ReadCluster)
         # self.distanceCache = DistanceCache(self.multiplicativeDistance)
-        self.distanceCache = DistanceCache(
-            self.commonNucleotidesAgreementDistance)
+        self.distanceCache = DistanceCache(self.commonNucleotidesAgreementDistance)
 
-    def __len__(self):
+    def __len__(self) -> int:
         """
         Our length is the number of read clusters.
 
@@ -193,7 +216,7 @@ class ReadClusters(object):
         """
         return len(self.readClusters)
 
-    def add(self, read):
+    def add(self, read: Read) -> int:
         """
         Add a single read as a new cluster.
 
@@ -204,11 +227,12 @@ class ReadClusters(object):
         self.readClusters[count].add(read)
         return count
 
-    def commonNucleotidesAgreementDistance(self, a, b):
+    def commonNucleotidesAgreementDistance(self, a: int, b: int) -> float:
         """
         Measure the distance from one cluster to another based on the fraction
         of shared offsets where the commonest nucleotide(s) for the offset
-        has a non-empty intersection. Hard to explain...
+        has a non-empty intersection. Pathetic docstring meta-comment: This is
+        hard to explain :-(
 
         @param a: An C{int} cluster number.
         @param b: An C{int} cluster number.
@@ -219,11 +243,13 @@ class ReadClusters(object):
 
         return 1.0 - (
             (1.0 - ReadCluster.commonNucleotidesAgreementDistance(a, b))
-            *
-            max(self.COMMON_OFFSETS_MAX_FRACTION_MIN,
-                ReadCluster.commonOffsetsMaxFraction(a, b)))
+            * max(
+                self.COMMON_OFFSETS_MAX_FRACTION_MIN,
+                ReadCluster.commonOffsetsMaxFraction(a, b),
+            )
+        )
 
-    def multiplicativeDistance(self, a, b):
+    def multiplicativeDistance(self, a: int, b: int) -> float:
         """
         Calculate an inter-cluster distance based on the maximum fraction of
         offsets (for the two clusters) present in common and the
@@ -238,11 +264,13 @@ class ReadClusters(object):
         # COMMON_OFFSETS_MAX_FRACTION_MIN.
         return 1.0 - (
             (1.0 - ReadCluster.commonNucleotidesMultiplicativeDistance(a, b))
-            *
-            max(self.COMMON_OFFSETS_MAX_FRACTION_MIN,
-                ReadCluster.commonOffsetsMaxFraction(a, b)))
+            * max(
+                self.COMMON_OFFSETS_MAX_FRACTION_MIN,
+                ReadCluster.commonOffsetsMaxFraction(a, b),
+            )
+        )
 
-    def mergeDescription(self, a, b, distance):
+    def mergeDescription(self, a: int, b: int, distance: float) -> str:
         """
         Make a textual description of a cluster merge.
 
@@ -259,73 +287,85 @@ class ReadClusters(object):
         matches = []
         sharedCount = matchCount = 0
 
-        allOffsets = sorted(
-            set(cluster1.nucleotides) | set(cluster2.nucleotides))
+        allOffsets = sorted(set(cluster1.nucleotides) | set(cluster2.nucleotides))
 
         for offset in allOffsets:
-
             inCount = 0
 
             if offset in cluster1.nucleotides:
                 result1.append(cluster1.nucleotides[offset].baseCountsToStr())
                 inCount += 1
             else:
-                result1.append('-')
+                result1.append("-")
 
             if offset in cluster2.nucleotides:
                 result2.append(cluster2.nucleotides[offset].baseCountsToStr())
                 inCount += 1
             else:
-                result2.append('-')
+                result2.append("-")
 
             if inCount == 2:
                 sharedCount += 1
-                if (cluster1.nucleotides[offset].commonest &
-                        cluster2.nucleotides[offset].commonest):
-                    matches.append('*')
+                if (
+                    cluster1.nucleotides[offset].commonest
+                    & cluster2.nucleotides[offset].commonest
+                ):
+                    matches.append("*")
                     matchCount += 1
                 else:
                     multiple = OffsetBases.highestFrequenciesMultiple(
-                        cluster1.nucleotides[offset],
-                        cluster2.nucleotides[offset])
+                        cluster1.nucleotides[offset], cluster2.nucleotides[offset]
+                    )
                     # Sanity: the multiple cannot be None because that
                     # would mean only one nucleotide is present, and that
                     # case is dealt with by the first part of this if/then.
                     assert multiple is not None
                     if multiple >= ReadCluster.MIN_COMMONEST_MULTIPLE:
                         matchCount += 1
-                        matches.append('+')
+                        matches.append("+")
                     else:
-                        matches.append('')
+                        matches.append("")
             else:
-                matches.append('')
+                matches.append("")
 
         result1Width = max(len(line) for line in result1)
         result2Width = max(len(line) for line in result2)
 
-        return '\n'.join(
+        return "\n".join(
             [
-                ('Merging clusters %d and %d with distance %.2f' %
-                 (a, b, distance)),
-                ('Cluster %d has %d read%s, covering %d offset%s' %
-                 (a,
-                  len(cluster1.reads), s(len(cluster1.reads), ),
-                  len(cluster1.nucleotides), s(len(cluster1.nucleotides)))),
-                ('Cluster %d has %d read%s, covering %d offset%s' %
-                 (b,
-                  len(cluster2.reads), s(len(cluster2.reads)),
-                  len(cluster2.nucleotides), s(len(cluster2.nucleotides)))),
-                ('%d matches out of %d shared offsets' %
-                 (matchCount, sharedCount)),
-            ] + [
-                '  %d: %*s    %*s    %s' %
-                (offset + 1, result1Width, line1, result2Width, line2, match)
-                for (offset, line1, line2, match) in
-                zip(allOffsets, result1, result2, matches)
+                ("Merging clusters %d and %d with distance %.2f" % (a, b, distance)),
+                (
+                    "Cluster %d has %d read%s, covering %d offset%s"
+                    % (
+                        a,
+                        len(cluster1.reads),
+                        s(len(cluster1.reads)),
+                        len(cluster1.nucleotides),
+                        s(len(cluster1.nucleotides)),
+                    )
+                ),
+                (
+                    "Cluster %d has %d read%s, covering %d offset%s"
+                    % (
+                        b,
+                        len(cluster2.reads),
+                        s(len(cluster2.reads)),
+                        len(cluster2.nucleotides),
+                        s(len(cluster2.nucleotides)),
+                    )
+                ),
+                ("%d matches out of %d shared offsets" % (matchCount, sharedCount)),
+            ]
+            + [
+                "  %d: %*s    %*s    %s"
+                % (offset + 1, result1Width, line1, result2Width, line2, match)
+                for (offset, line1, line2, match) in zip(
+                    allOffsets, result1, result2, matches
+                )
             ]
         )
 
-    def mergeDescriptionWithOffsetScores(self, a, b, distance):
+    def mergeDescriptionWithOffsetScores(self, a: int, b: int, distance: float) -> str:
         """
         Make a textual description of a cluster merge, including per-offset
         score information.
@@ -344,76 +384,103 @@ class ReadClusters(object):
         matches = []
         sharedCount = matchCount = 0
 
-        allOffsets = sorted(
-            set(cluster1.nucleotides) | set(cluster2.nucleotides))
+        allOffsets = sorted(set(cluster1.nucleotides) | set(cluster2.nucleotides))
 
         for offset in allOffsets:
-
             inCount = 0
 
             if offset in cluster1.nucleotides:
                 result1.append(cluster1.nucleotides[offset].baseCountsToStr())
                 inCount += 1
             else:
-                result1.append('-')
+                result1.append("-")
 
             if offset in cluster2.nucleotides:
                 result2.append(cluster2.nucleotides[offset].baseCountsToStr())
                 inCount += 1
             else:
-                result2.append('-')
+                result2.append("-")
 
             if inCount == 2:
                 sharedCount += 1
-                if (cluster1.nucleotides[offset].commonest &
-                        cluster2.nucleotides[offset].commonest):
-                    matches.append('*')
+                if (
+                    cluster1.nucleotides[offset].commonest
+                    & cluster2.nucleotides[offset].commonest
+                ):
+                    matches.append("*")
                     matchCount += 1
                 else:
-                    matches.append('')
+                    matches.append("")
 
                 offsetScores.append(
-                    '%.3f' % min(
+                    "%.3f"
+                    % min(
                         OffsetBases.multiplicativeDistance(
-                            cluster1.nucleotides[offset],
-                            cluster2.nucleotides[offset]),
+                            cluster1.nucleotides[offset], cluster2.nucleotides[offset]
+                        ),
                         OffsetBases.homogeneousDistance(
-                            cluster1.nucleotides[offset],
-                            cluster2.nucleotides[offset])))
+                            cluster1.nucleotides[offset], cluster2.nucleotides[offset]
+                        ),
+                    )
+                )
             else:
-                matches.append('')
-                offsetScores.append('')
+                matches.append("")
+                offsetScores.append("")
 
         result1Width = max(len(line) for line in result1)
         result2Width = max(len(line) for line in result2)
         offsetScoresWidth = max(len(line) for line in offsetScores)
 
-        return '\n'.join(
+        return "\n".join(
             [
-                ('Merging clusters %d and %d with distance %.2f' %
-                 (a, b, distance)),
-                ('Cluster %d has %d read%s, covering %d offset%s' %
-                 (a,
-                  len(cluster1.reads), s(len(cluster1.reads), ),
-                  len(cluster1.nucleotides), s(len(cluster1.nucleotides)))),
-                ('Cluster %d has %d read%s, covering %d offset%s' %
-                 (b,
-                  len(cluster2.reads), s(len(cluster2.reads)),
-                  len(cluster2.nucleotides), s(len(cluster2.nucleotides)))),
-                ('%d matches out of %d shared offsets' %
-                 (matchCount, sharedCount)),
-            ] + [
-                '  %d: %*s    %*s    %*s    %s' %
-                (offset + 1, result1Width, line1, result2Width, line2,
-                 offsetScoresWidth, offsetScore, match)
-                for (offset, line1, line2, offsetScore, match) in
-                zip(allOffsets, result1, result2, offsetScores, matches)
+                ("Merging clusters %d and %d with distance %.2f" % (a, b, distance)),
+                (
+                    "Cluster %d has %d read%s, covering %d offset%s"
+                    % (
+                        a,
+                        len(cluster1.reads),
+                        s(
+                            len(cluster1.reads),
+                        ),
+                        len(cluster1.nucleotides),
+                        s(len(cluster1.nucleotides)),
+                    )
+                ),
+                (
+                    "Cluster %d has %d read%s, covering %d offset%s"
+                    % (
+                        b,
+                        len(cluster2.reads),
+                        s(len(cluster2.reads)),
+                        len(cluster2.nucleotides),
+                        s(len(cluster2.nucleotides)),
+                    )
+                ),
+                ("%d matches out of %d shared offsets" % (matchCount, sharedCount)),
+            ]
+            + [
+                "  %d: %*s    %*s    %*s    %s"
+                % (
+                    offset + 1,
+                    result1Width,
+                    line1,
+                    result2Width,
+                    line2,
+                    offsetScoresWidth,
+                    offsetScore,
+                    match,
+                )
+                for (offset, line1, line2, offsetScore, match) in zip(
+                    allOffsets, result1, result2, offsetScores, matches
+                )
             ]
         )
 
-    def analyze(self, cutoff, fp=None):
+    def analyze(
+        self, cutoff: float, fp: Optional[TextIO] = None
+    ) -> ValuesView[ReadCluster]:
         """
-        Perform the cluster analysis, up to a given distance cut off.
+        Perform the cluster analysis, up to a given distance cut-off.
 
         @param cutoff: The C{float} distance at which clustering will be
             stopped. Clusters at any greater distance will not be merged.
@@ -424,7 +491,7 @@ class ReadClusters(object):
             merged.
         """
         if fp:
-            print('Starting cluster analysis on %d reads' % len(self), file=fp)
+            print("Starting cluster analysis on %d reads" % len(self), file=fp)
 
         for a in self.readClusters:
             self.distanceCache.add(a)
@@ -441,7 +508,7 @@ class ReadClusters(object):
             if fp:
                 print(self.mergeDescription(a, b, lowestDistance), file=fp)
 
-            # Merge.
+            # Merge the clusters.
             count = next(self._count)
             self.readClusters[a].merge(self.readClusters[b])
             self.readClusters[count] = self.readClusters[a]
@@ -452,26 +519,40 @@ class ReadClusters(object):
             del self.readClusters[b]
 
             if fp:
-                print('New cluster (%d) has %d reads and covers %d offsets. '
-                      'Remaining cluster count: %d.' %
-                      (count, len(self.readClusters[count].reads),
-                       len(self.readClusters[count].nucleotides), len(self)),
-                      file=fp)
+                print(
+                    "New cluster (%d) has %d reads and covers %d offsets. "
+                    "Remaining cluster count: %d."
+                    % (
+                        count,
+                        len(self.readClusters[count].reads),
+                        len(self.readClusters[count].nucleotides),
+                        len(self),
+                    ),
+                    file=fp,
+                )
 
         if fp:
             if lowestDistance is None:
-                print('Distance priority queue is empty, all reads '
-                      'are now in one cluster.', file=fp)
+                print(
+                    "Distance priority queue is empty, all reads "
+                    "are now in one cluster.",
+                    file=fp,
+                )
             else:
-                print('The lowest inter-cluster distance has risen '
-                      'to %.2f, which exceeds the %.2f cutoff value. '
-                      '%d unmerged clusters remain.' %
-                      (lowestDistance, cutoff, len(self.readClusters)),
-                      file=fp)
+                print(
+                    "The lowest inter-cluster distance has risen "
+                    "to %.2f, which exceeds the %.2f cutoff value. "
+                    "%d unmerged clusters remain."
+                    % (lowestDistance, cutoff, len(self.readClusters)),
+                    file=fp,
+                )
 
                 # Print the remaining inter-cluster distances.
-                print('\nThe remaining cluster differences that were too '
-                      'distant to qualify are:', file=fp)
+                print(
+                    "\nThe remaining cluster differences that were too "
+                    "distant to qualify are:",
+                    file=fp,
+                )
 
                 while True:
                     lowestDistance = self.distanceCache.lowestDistance()
@@ -479,12 +560,13 @@ class ReadClusters(object):
                         break
 
                     a, b = self.distanceCache.pop()
-                    # self.distanceCache.remove(a)
-                    # self.distanceCache.remove(b)
+                    # 2024-03-30 why were the next two lines commented out?
+                    self.distanceCache.remove(a)
+                    self.distanceCache.remove(b)
 
                     print(self.mergeDescription(a, b, lowestDistance), file=fp)
 
         if fp:
-            print('End of cluster analysis.', file=fp)
+            print("End of cluster analysis.", file=fp)
 
         return self.readClusters.values()
